#include <fstream>
#include <iostream>
#include <stack>
#include <vector>
#include "sdsl/lcp.hpp"
#include "sdsl/suffix_arrays.hpp"

using namespace std;

sdsl::int_vector<8> T;          // text
size_t N;                       // size
sdsl::csa_bitcompressed<> SA;   // suffix-array
sdsl::lcp_bitcompressed<> LCP;  // LCP-array
vector<int> LPF;                // LPF-array -> will be overwritten by LPnF-values
vector<int> prevOcc;            // prevOcc-array -> if not needed it can be omitted

struct lcp_interval {
    size_t lcp, lb, idx;
    vector<int> SAvalues;
    lcp_interval(size_t lcp, size_t lb, size_t idx, vector<int> SAvalues)
        : lcp(lcp), lb(lb), idx(idx), SAvalues(SAvalues){};
};

int process(vector<int> &SAvalues, int lcp);

void bottom_up_traversal() {
    lcp_interval *lastInterval = NULL;
    int m = 0;
    stack<lcp_interval> sta;
    sta.push(lcp_interval(0, 0, 0, vector<int>()));
    for (size_t k = 1; k <= N; k++) {
        size_t lb = k - 1;
        while (LCP[k] < sta.top().lcp) {
            lastInterval = &sta.top();
            if (lastInterval->idx == k - 1) {
                lastInterval->SAvalues.push_back(SA[k - 1]);
            }
            m = process(lastInterval->SAvalues, lastInterval->lcp);
            lb = lastInterval->lb;
            sta.pop();
            if (LCP[k] <= sta.top().lcp) {
                sta.top().SAvalues.push_back(m);
                lastInterval = NULL;
            }
        }
        if (LCP[k] > sta.top().lcp) {
            if (lastInterval) {
                sta.push(lcp_interval(LCP[k], lb, k, vector<int>{m}));
                lastInterval = NULL;
            } else {
                sta.push(lcp_interval(LCP[k], lb, k, vector<int>()));
            }
            if (lb == k - 1) {
                sta.top().SAvalues.push_back(SA[lb]);
            }
        } else {
            if (sta.top().idx == k - 1) {
                sta.top().SAvalues.push_back(SA[k - 1]);
            }
            sta.top().idx = k;
        }
    }
    // prevOcc- and LPF-entries are already 0, hence following code could be
    // omitted
    lastInterval = &sta.top();
    for (int i : lastInterval->SAvalues) {
        prevOcc[i] = 0;
        LPF[i] = 0;
    }
    sta.pop();
}

int process(vector<int> &SAvalues, int lcp) {
    int m = SAvalues[0];
    for (int i : SAvalues) {
        if (i < m) m = i;
    }
    for (int i : SAvalues) {
        if (i != m) {
            prevOcc[i] = m;
            LPF[i] = lcp;
        }
    }
    return m;
}

void computeLPnF() {  // modified algorithm 1 stated in the paper
    for (int i = N - 1; i >= 0; i--) {
        if (LPF[i] > 0) {
            int j = prevOcc[i];
            if (j + LPF[i] > i) {
                int l = i - j;
                if (LPF[j] > l) {
                    prevOcc[i] = prevOcc[j];
                    LPF[i] = LPF[j];
                } else {
                    prevOcc[i] = j;
                    LPF[i] = l;
                }
            }
        }
    }
}

int main(int argc, char *argv[]) {
    if (2 <= argc && argc <= 4) {
        sdsl::cache_config config(false);
        sdsl::construct(SA, argv[1], config, 1);
        sdsl::load_from_cache(T, sdsl::key_text_trait<8>::KEY_TEXT, config);

        config.delete_files = true;
        sdsl::construct(LCP, argv[1], config, 1);

        N = SA.F.size();
        LPF = vector<int>(N);
        prevOcc = vector<int>(N);
        bottom_up_traversal();  // produces leftmost LPF-array entries

        N--;
        computeLPnF();

        // f-factor output
        if (3 <= argc && argc <= 4) {
            ofstream LPnF_out(argv[2], ifstream::out);
            for (size_t i = 0; i < N; i++) {
                LPnF_out << LPF[i] << " ";
#ifdef f_factor
                if (LPF[i] > 1) {
                    i += LPF[i] - 1;
                }
#endif
            }
            LPnF_out << endl;
            LPnF_out.close();
        }

        // prevOcc output
        if (argc == 4) {
            ofstream prevOcc_out(argv[3], ifstream::out);
            for (size_t i = 0; i < N; i++) {
                if (LPF[i] == 0) {
                    prevOcc_out << (int)T[i] << " ";
                } else {
                    prevOcc_out << prevOcc[i] << " ";
                }
#ifdef f_factor
                if (LPF[i] > 1) {
                    i += LPF[i] - 1;
                }
#endif
            }
            prevOcc_out << endl;
            prevOcc_out.close();
        }
    } else {
        cerr << "usage: " << argv[0]
             << " <input-filepath> [<LPnF/f-factor_path>] [<prevOcc_path>]"
             << endl;
        return EXIT_FAILURE;
    }
}
