#include <assert.h>
#include <cstring>
#include <fstream>
#include <iostream>
#include <sstream>
#include <string>
#include "divsufsort.h"

using namespace std;

string T;      // text
long long N;   // size
int *SA;       // suffix-array -> will be overwritten by LPnF-values
int *prevOcc;  // prevOcc-array -> if not needed it can be omitted

//// CODE BY SIMON GOG --> ////
void sop(int i, int l, int j, int *lps, int *prevOcc, int bot) {
    if (j == 0 and l == 0 and i == 0) return;
    assert(i > j);
    assert(i != j);
    if (lps[i] == bot) {
        lps[i] = l;
        prevOcc[i] = j;
    } else {
        if (lps[i] < l) {
            if (prevOcc[i] > j)
                sop(prevOcc[i], lps[i], j, lps, prevOcc, bot);
            else
                sop(j, lps[i], prevOcc[i], lps, prevOcc, bot);
            lps[i] = l;
            prevOcc[i] = j;
        } else {
            if (prevOcc[i] > j)
                sop(prevOcc[i], l, j, lps, prevOcc, bot);
            else
                sop(j, l, prevOcc[i], lps, prevOcc, bot);
        }
    }
}

void calcLPF() {
    for (int i = 0; i < N; i++) prevOcc[i] = -1;

    // compute LZ factorization
    // compute PHI
    int *phi = new int[N];

    long long to_add[2] = {-1, N - 1};
    for (int i = 0; i < N; i++) {
        phi[SA[i]] = SA[i + to_add[i == 0]];
    }
    // SA holds now LPS
    for (int i = 0; i < N; ++i) SA[i] = -1;

    int maxFactorLength = 0;
    int l = 0;
    for (int i = 0; i < N; ++i) {
        int j = phi[i];
        while (T[i + l] == T[j + l]) ++l;
        if (i > j) {
            sop(i, l, j, SA, prevOcc, -1);
        } else {
            sop(j, l, i, SA, prevOcc, -1);
        }
        if (l > 0) --l;
    }
    int numfactors = 0;
    int longFactors = 0;
    float averageFactorLength = 0;
    SA[0] = 0;
    for (int i = 0; i < N;) {
        ++numfactors;
        if (SA[i] > maxFactorLength) maxFactorLength = SA[i];
        if (SA[i] > 8) ++longFactors;
        if (SA[i] < 1) {
            averageFactorLength += 1;
            ++i;
        } else {
            averageFactorLength += SA[i];
            i += SA[i];
        }
    }
    delete[] phi;
}
//// <-- CODE BY SIMON GOG ////

void compute_lmOcc() {  // algorithm 2 stated in the paper
    for (int i = 0; i < N - 1; i++) {
        if (SA[i] > 0) {
            int j = prevOcc[i];
            while (SA[j] >= SA[i]) {
                j = prevOcc[j];
            }
            prevOcc[i] = j;
        }
    }
}

void computeLPnF_modified() {  // modified algorithm 1 stated in the paper
    for (int i = N - 1; i >= 0; i--) {
        if (SA[i] > 0) {
            int j = prevOcc[i];
            if (j + SA[i] > i) {
                int l = i - j;
                if (SA[j] > l) {
                    prevOcc[i] = prevOcc[j];
                    SA[i] = SA[j];
                } else {
                    prevOcc[i] = j;
                    SA[i] = l;
                }
            }
        }
    }
}

void computeLPnF() {  // algorithm 1 stated in the paper
    for (int i = N - 1; i >= 0; i--) {
        if (SA[i] > 0) {
            int j = prevOcc[i];
            if (j + SA[i] > i) {
                int l = i - j;
                while (SA[j] > l) {
                    l = min(SA[i], SA[j]);
                    j = prevOcc[j];
                    if (j + l <= i) {
                        break;
                    } else {
                        l = i - j;
                    }
                }
                prevOcc[i] = j;
                SA[i] = l;
            }
        }
    }
}

void loadFile(char *fname) {
    ifstream ifs(fname, ifstream::in);
    if (ifs.fail()) {
        cerr << "Could not open file " << fname << endl;
        exit(EXIT_FAILURE);
    }
    stringstream ss;
    ss << ifs.rdbuf();
    T = ss.str();
    ifs.close();

    N = T.length();
    if (N <= 0) exit(0);
}

int main(int argc, char *argv[]) {
    if (3 <= argc && argc <= 5) {
        loadFile(argv[1]);

        SA = new int[N];
        divsufsort((unsigned char *)&T[0], SA, N);

        prevOcc = new int[N];

        calcLPF();

        if (strcmp(argv[2], "1") == 0) {
            compute_lmOcc();         // calculate leftmost prevOcc
            computeLPnF_modified();  // calculate LPnF values
        } else if (strcmp(argv[2], "0") == 0) {
            computeLPnF();  // calculate LPnF values from a non leftmost
                            // prevOcc-array
        } else {
            cerr << "usage: " << argv[0] << " <filepath> <with-alg2=0/1> "
                 << "[<LPnF/f-factor_path>] [<prev-occ_path>]" << endl;
            return EXIT_FAILURE;
        }

        // f-factor output
        if (4 <= argc && argc <= 5) {
            ofstream LPnF_out(argv[3], ifstream::out);
            for (int i = 0; i < N; i++) {
                LPnF_out << SA[i] << " ";
#ifdef f_factor
                if (SA[i] > 1) {
                    i += SA[i] - 1;
                }
#endif
            }
            LPnF_out << endl;
            LPnF_out.close();
        }

        // prevOcc output
        if (argc == 5) {
            ofstream prevOcc_out(argv[4], ifstream::out);
            for (int i = 0; i < N; i++) {
                if (SA[i] == 0) {
                    prevOcc_out << T[i] << " ";
                } else {
                    prevOcc_out << prevOcc[i] << " ";
                }
#ifdef f_factor
                if (SA[i] > 1) {
                    i += SA[i] - 1;
                }
#endif
            }
            prevOcc_out << endl;
            prevOcc_out.close();
        }

        delete[] SA;
        delete[] prevOcc;
    } else {
        cerr << "usage: " << argv[0] << " <filepath> <with-alg2=0/1> "
             << "[<LPnF/f-factor_path>] [<prev-occ_path>]" << endl;
        return EXIT_FAILURE;
    }
}
